// Copyright (c) 2012- PPSSPP Project.

// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, version 2.0 or later versions.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License 2.0 for more details.

// A copy of the GPL 2.0 should have been included with the program.
// If not, see http://www.gnu.org/licenses/

// Official git repository and contact information can be found at
// https://github.com/hrydgard/ppsspp and http://www.ppsspp.org/.

#include "Common/Serialize/Serializer.h"
#include "Common/Serialize/SerializeFuncs.h"
#include "Core/FileSystems/FileSystem.h"

void PSPFileInfo::DoState(PointerWrap &p) {
	auto s = p.Section("PSPFileInfo", 1);
	if (!s)
		return;

	Do(p, name);
	Do(p, size);
	Do(p, access);
	Do(p, exists);
	Do(p, type);
	Do(p, atime);
	Do(p, ctime);
	Do(p, mtime);
	Do(p, isOnSectorSystem);
	Do(p, startSector);
	Do(p, numSectors);
	Do(p, sectorSize);
}

