/****************************************************************************
** Meta object code from reading C++ file 'mainwindow.h'
**
** Created by: The Qt Meta Object Compiler version 67 (Qt 5.15.18)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include <memory>
#include "../../../ppsspp/Qt/mainwindow.h"
#include <QtCore/qbytearray.h>
#include <QtCore/qmetatype.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'mainwindow.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 67
#error "This file was generated using the moc from 5.15.18. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

QT_BEGIN_MOC_NAMESPACE
QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
struct qt_meta_stringdata_MainWindow_t {
    QByteArrayData data[65];
    char stringdata0[920];
};
#define QT_MOC_LITERAL(idx, ofs, len) \
    Q_STATIC_BYTE_ARRAY_DATA_HEADER_INITIALIZER_WITH_OFFSET(len, \
    qptrdiff(offsetof(qt_meta_stringdata_MainWindow_t, stringdata0) + ofs \
        - idx * sizeof(QByteArrayData)) \
    )
static const qt_meta_stringdata_MainWindow_t qt_meta_stringdata_MainWindow = {
    {
QT_MOC_LITERAL(0, 0, 10), // "MainWindow"
QT_MOC_LITERAL(1, 11, 11), // "retranslate"
QT_MOC_LITERAL(2, 23, 0), // ""
QT_MOC_LITERAL(3, 24, 10), // "updateMenu"
QT_MOC_LITERAL(4, 35, 8), // "newFrame"
QT_MOC_LITERAL(5, 44, 7), // "loadAct"
QT_MOC_LITERAL(6, 52, 8), // "closeAct"
QT_MOC_LITERAL(7, 61, 9), // "openmsAct"
QT_MOC_LITERAL(8, 71, 24), // "saveStateGroup_triggered"
QT_MOC_LITERAL(9, 96, 8), // "QAction*"
QT_MOC_LITERAL(10, 105, 6), // "action"
QT_MOC_LITERAL(11, 112, 10), // "qlstateAct"
QT_MOC_LITERAL(12, 123, 10), // "qsstateAct"
QT_MOC_LITERAL(13, 134, 9), // "lstateAct"
QT_MOC_LITERAL(14, 144, 9), // "sstateAct"
QT_MOC_LITERAL(15, 154, 16), // "recordDisplayAct"
QT_MOC_LITERAL(16, 171, 24), // "useLosslessVideoCodecAct"
QT_MOC_LITERAL(17, 196, 18), // "useOutputBufferAct"
QT_MOC_LITERAL(18, 215, 14), // "recordAudioAct"
QT_MOC_LITERAL(19, 230, 7), // "exitAct"
QT_MOC_LITERAL(20, 238, 6), // "runAct"
QT_MOC_LITERAL(21, 245, 8), // "pauseAct"
QT_MOC_LITERAL(22, 254, 7), // "stopAct"
QT_MOC_LITERAL(23, 262, 8), // "resetAct"
QT_MOC_LITERAL(24, 271, 12), // "switchUMDAct"
QT_MOC_LITERAL(25, 284, 30), // "displayRotationGroup_triggered"
QT_MOC_LITERAL(26, 315, 14), // "breakonloadAct"
QT_MOC_LITERAL(27, 330, 16), // "ignoreIllegalAct"
QT_MOC_LITERAL(28, 347, 7), // "lmapAct"
QT_MOC_LITERAL(29, 355, 7), // "smapAct"
QT_MOC_LITERAL(30, 363, 7), // "lsymAct"
QT_MOC_LITERAL(31, 371, 7), // "ssymAct"
QT_MOC_LITERAL(32, 379, 13), // "resetTableAct"
QT_MOC_LITERAL(33, 393, 11), // "dumpNextAct"
QT_MOC_LITERAL(34, 405, 10), // "takeScreen"
QT_MOC_LITERAL(35, 416, 10), // "consoleAct"
QT_MOC_LITERAL(36, 427, 11), // "languageAct"
QT_MOC_LITERAL(37, 439, 17), // "controlMappingAct"
QT_MOC_LITERAL(38, 457, 22), // "displayLayoutEditorAct"
QT_MOC_LITERAL(39, 480, 15), // "moreSettingsAct"
QT_MOC_LITERAL(40, 496, 15), // "bufferRenderAct"
QT_MOC_LITERAL(41, 512, 9), // "linearAct"
QT_MOC_LITERAL(42, 522, 34), // "renderingResolutionGroup_trig..."
QT_MOC_LITERAL(43, 557, 21), // "windowGroup_triggered"
QT_MOC_LITERAL(44, 579, 16), // "autoframeskipAct"
QT_MOC_LITERAL(45, 596, 28), // "frameSkippingGroup_triggered"
QT_MOC_LITERAL(46, 625, 31), // "textureFilteringGroup_triggered"
QT_MOC_LITERAL(47, 657, 34), // "screenScalingFilterGroup_trig..."
QT_MOC_LITERAL(48, 692, 34), // "textureScalingLevelGroup_trig..."
QT_MOC_LITERAL(49, 727, 33), // "textureScalingTypeGroup_trigg..."
QT_MOC_LITERAL(50, 761, 14), // "deposterizeAct"
QT_MOC_LITERAL(51, 776, 12), // "transformAct"
QT_MOC_LITERAL(52, 789, 12), // "frameskipAct"
QT_MOC_LITERAL(53, 802, 8), // "audioAct"
QT_MOC_LITERAL(54, 811, 9), // "cheatsAct"
QT_MOC_LITERAL(55, 821, 7), // "chatAct"
QT_MOC_LITERAL(56, 829, 10), // "fullscrAct"
QT_MOC_LITERAL(57, 840, 12), // "raiseTopMost"
QT_MOC_LITERAL(58, 853, 10), // "websiteAct"
QT_MOC_LITERAL(59, 864, 8), // "forumAct"
QT_MOC_LITERAL(60, 873, 7), // "goldAct"
QT_MOC_LITERAL(61, 881, 6), // "gitAct"
QT_MOC_LITERAL(62, 888, 10), // "discordAct"
QT_MOC_LITERAL(63, 899, 8), // "aboutAct"
QT_MOC_LITERAL(64, 908, 11) // "langChanged"

    },
    "MainWindow\0retranslate\0\0updateMenu\0"
    "newFrame\0loadAct\0closeAct\0openmsAct\0"
    "saveStateGroup_triggered\0QAction*\0"
    "action\0qlstateAct\0qsstateAct\0lstateAct\0"
    "sstateAct\0recordDisplayAct\0"
    "useLosslessVideoCodecAct\0useOutputBufferAct\0"
    "recordAudioAct\0exitAct\0runAct\0pauseAct\0"
    "stopAct\0resetAct\0switchUMDAct\0"
    "displayRotationGroup_triggered\0"
    "breakonloadAct\0ignoreIllegalAct\0lmapAct\0"
    "smapAct\0lsymAct\0ssymAct\0resetTableAct\0"
    "dumpNextAct\0takeScreen\0consoleAct\0"
    "languageAct\0controlMappingAct\0"
    "displayLayoutEditorAct\0moreSettingsAct\0"
    "bufferRenderAct\0linearAct\0"
    "renderingResolutionGroup_triggered\0"
    "windowGroup_triggered\0autoframeskipAct\0"
    "frameSkippingGroup_triggered\0"
    "textureFilteringGroup_triggered\0"
    "screenScalingFilterGroup_triggered\0"
    "textureScalingLevelGroup_triggered\0"
    "textureScalingTypeGroup_triggered\0"
    "deposterizeAct\0transformAct\0frameskipAct\0"
    "audioAct\0cheatsAct\0chatAct\0fullscrAct\0"
    "raiseTopMost\0websiteAct\0forumAct\0"
    "goldAct\0gitAct\0discordAct\0aboutAct\0"
    "langChanged"
};
#undef QT_MOC_LITERAL

static const uint qt_meta_data_MainWindow[] = {

 // content:
       8,       // revision
       0,       // classname
       0,    0, // classinfo
      61,   14, // methods
       0,    0, // properties
       0,    0, // enums/sets
       0,    0, // constructors
       0,       // flags
       2,       // signalCount

 // signals: name, argc, parameters, tag, flags
       1,    0,  319,    2, 0x06 /* Public */,
       3,    0,  320,    2, 0x06 /* Public */,

 // slots: name, argc, parameters, tag, flags
       4,    0,  321,    2, 0x0a /* Public */,
       5,    0,  322,    2, 0x08 /* Private */,
       6,    0,  323,    2, 0x08 /* Private */,
       7,    0,  324,    2, 0x08 /* Private */,
       8,    1,  325,    2, 0x08 /* Private */,
      11,    0,  328,    2, 0x08 /* Private */,
      12,    0,  329,    2, 0x08 /* Private */,
      13,    0,  330,    2, 0x08 /* Private */,
      14,    0,  331,    2, 0x08 /* Private */,
      15,    0,  332,    2, 0x08 /* Private */,
      16,    0,  333,    2, 0x08 /* Private */,
      17,    0,  334,    2, 0x08 /* Private */,
      18,    0,  335,    2, 0x08 /* Private */,
      19,    0,  336,    2, 0x08 /* Private */,
      20,    0,  337,    2, 0x08 /* Private */,
      21,    0,  338,    2, 0x08 /* Private */,
      22,    0,  339,    2, 0x08 /* Private */,
      23,    0,  340,    2, 0x08 /* Private */,
      24,    0,  341,    2, 0x08 /* Private */,
      25,    1,  342,    2, 0x08 /* Private */,
      26,    0,  345,    2, 0x08 /* Private */,
      27,    0,  346,    2, 0x08 /* Private */,
      28,    0,  347,    2, 0x08 /* Private */,
      29,    0,  348,    2, 0x08 /* Private */,
      30,    0,  349,    2, 0x08 /* Private */,
      31,    0,  350,    2, 0x08 /* Private */,
      32,    0,  351,    2, 0x08 /* Private */,
      33,    0,  352,    2, 0x08 /* Private */,
      34,    0,  353,    2, 0x08 /* Private */,
      35,    0,  354,    2, 0x08 /* Private */,
      36,    0,  355,    2, 0x08 /* Private */,
      37,    0,  356,    2, 0x08 /* Private */,
      38,    0,  357,    2, 0x08 /* Private */,
      39,    0,  358,    2, 0x08 /* Private */,
      40,    0,  359,    2, 0x08 /* Private */,
      41,    0,  360,    2, 0x08 /* Private */,
      42,    1,  361,    2, 0x08 /* Private */,
      43,    1,  364,    2, 0x08 /* Private */,
      44,    0,  367,    2, 0x08 /* Private */,
      45,    1,  368,    2, 0x08 /* Private */,
      46,    1,  371,    2, 0x08 /* Private */,
      47,    1,  374,    2, 0x08 /* Private */,
      48,    1,  377,    2, 0x08 /* Private */,
      49,    1,  380,    2, 0x08 /* Private */,
      50,    0,  383,    2, 0x08 /* Private */,
      51,    0,  384,    2, 0x08 /* Private */,
      52,    0,  385,    2, 0x08 /* Private */,
      53,    0,  386,    2, 0x08 /* Private */,
      54,    0,  387,    2, 0x08 /* Private */,
      55,    0,  388,    2, 0x08 /* Private */,
      56,    0,  389,    2, 0x08 /* Private */,
      57,    0,  390,    2, 0x08 /* Private */,
      58,    0,  391,    2, 0x08 /* Private */,
      59,    0,  392,    2, 0x08 /* Private */,
      60,    0,  393,    2, 0x08 /* Private */,
      61,    0,  394,    2, 0x08 /* Private */,
      62,    0,  395,    2, 0x08 /* Private */,
      63,    0,  396,    2, 0x08 /* Private */,
      64,    1,  397,    2, 0x08 /* Private */,

 // signals: parameters
    QMetaType::Void,
    QMetaType::Void,

 // slots: parameters
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void, 0x80000000 | 9,   10,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void, 0x80000000 | 9,   10,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void, 0x80000000 | 9,   10,
    QMetaType::Void, 0x80000000 | 9,   10,
    QMetaType::Void,
    QMetaType::Void, 0x80000000 | 9,   10,
    QMetaType::Void, 0x80000000 | 9,   10,
    QMetaType::Void, 0x80000000 | 9,   10,
    QMetaType::Void, 0x80000000 | 9,   10,
    QMetaType::Void, 0x80000000 | 9,   10,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void, 0x80000000 | 9,   10,

       0        // eod
};

void MainWindow::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    if (_c == QMetaObject::InvokeMetaMethod) {
        auto *_t = static_cast<MainWindow *>(_o);
        (void)_t;
        switch (_id) {
        case 0: _t->retranslate(); break;
        case 1: _t->updateMenu(); break;
        case 2: _t->newFrame(); break;
        case 3: _t->loadAct(); break;
        case 4: _t->closeAct(); break;
        case 5: _t->openmsAct(); break;
        case 6: _t->saveStateGroup_triggered((*reinterpret_cast< QAction*(*)>(_a[1]))); break;
        case 7: _t->qlstateAct(); break;
        case 8: _t->qsstateAct(); break;
        case 9: _t->lstateAct(); break;
        case 10: _t->sstateAct(); break;
        case 11: _t->recordDisplayAct(); break;
        case 12: _t->useLosslessVideoCodecAct(); break;
        case 13: _t->useOutputBufferAct(); break;
        case 14: _t->recordAudioAct(); break;
        case 15: _t->exitAct(); break;
        case 16: _t->runAct(); break;
        case 17: _t->pauseAct(); break;
        case 18: _t->stopAct(); break;
        case 19: _t->resetAct(); break;
        case 20: _t->switchUMDAct(); break;
        case 21: _t->displayRotationGroup_triggered((*reinterpret_cast< QAction*(*)>(_a[1]))); break;
        case 22: _t->breakonloadAct(); break;
        case 23: _t->ignoreIllegalAct(); break;
        case 24: _t->lmapAct(); break;
        case 25: _t->smapAct(); break;
        case 26: _t->lsymAct(); break;
        case 27: _t->ssymAct(); break;
        case 28: _t->resetTableAct(); break;
        case 29: _t->dumpNextAct(); break;
        case 30: _t->takeScreen(); break;
        case 31: _t->consoleAct(); break;
        case 32: _t->languageAct(); break;
        case 33: _t->controlMappingAct(); break;
        case 34: _t->displayLayoutEditorAct(); break;
        case 35: _t->moreSettingsAct(); break;
        case 36: _t->bufferRenderAct(); break;
        case 37: _t->linearAct(); break;
        case 38: _t->renderingResolutionGroup_triggered((*reinterpret_cast< QAction*(*)>(_a[1]))); break;
        case 39: _t->windowGroup_triggered((*reinterpret_cast< QAction*(*)>(_a[1]))); break;
        case 40: _t->autoframeskipAct(); break;
        case 41: _t->frameSkippingGroup_triggered((*reinterpret_cast< QAction*(*)>(_a[1]))); break;
        case 42: _t->textureFilteringGroup_triggered((*reinterpret_cast< QAction*(*)>(_a[1]))); break;
        case 43: _t->screenScalingFilterGroup_triggered((*reinterpret_cast< QAction*(*)>(_a[1]))); break;
        case 44: _t->textureScalingLevelGroup_triggered((*reinterpret_cast< QAction*(*)>(_a[1]))); break;
        case 45: _t->textureScalingTypeGroup_triggered((*reinterpret_cast< QAction*(*)>(_a[1]))); break;
        case 46: _t->deposterizeAct(); break;
        case 47: _t->transformAct(); break;
        case 48: _t->frameskipAct(); break;
        case 49: _t->audioAct(); break;
        case 50: _t->cheatsAct(); break;
        case 51: _t->chatAct(); break;
        case 52: _t->fullscrAct(); break;
        case 53: _t->raiseTopMost(); break;
        case 54: _t->websiteAct(); break;
        case 55: _t->forumAct(); break;
        case 56: _t->goldAct(); break;
        case 57: _t->gitAct(); break;
        case 58: _t->discordAct(); break;
        case 59: _t->aboutAct(); break;
        case 60: _t->langChanged((*reinterpret_cast< QAction*(*)>(_a[1]))); break;
        default: ;
        }
    } else if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        switch (_id) {
        default: *reinterpret_cast<int*>(_a[0]) = -1; break;
        case 6:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<int*>(_a[0]) = -1; break;
            case 0:
                *reinterpret_cast<int*>(_a[0]) = qRegisterMetaType< QAction* >(); break;
            }
            break;
        case 21:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<int*>(_a[0]) = -1; break;
            case 0:
                *reinterpret_cast<int*>(_a[0]) = qRegisterMetaType< QAction* >(); break;
            }
            break;
        case 38:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<int*>(_a[0]) = -1; break;
            case 0:
                *reinterpret_cast<int*>(_a[0]) = qRegisterMetaType< QAction* >(); break;
            }
            break;
        case 39:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<int*>(_a[0]) = -1; break;
            case 0:
                *reinterpret_cast<int*>(_a[0]) = qRegisterMetaType< QAction* >(); break;
            }
            break;
        case 41:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<int*>(_a[0]) = -1; break;
            case 0:
                *reinterpret_cast<int*>(_a[0]) = qRegisterMetaType< QAction* >(); break;
            }
            break;
        case 42:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<int*>(_a[0]) = -1; break;
            case 0:
                *reinterpret_cast<int*>(_a[0]) = qRegisterMetaType< QAction* >(); break;
            }
            break;
        case 43:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<int*>(_a[0]) = -1; break;
            case 0:
                *reinterpret_cast<int*>(_a[0]) = qRegisterMetaType< QAction* >(); break;
            }
            break;
        case 44:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<int*>(_a[0]) = -1; break;
            case 0:
                *reinterpret_cast<int*>(_a[0]) = qRegisterMetaType< QAction* >(); break;
            }
            break;
        case 45:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<int*>(_a[0]) = -1; break;
            case 0:
                *reinterpret_cast<int*>(_a[0]) = qRegisterMetaType< QAction* >(); break;
            }
            break;
        case 60:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<int*>(_a[0]) = -1; break;
            case 0:
                *reinterpret_cast<int*>(_a[0]) = qRegisterMetaType< QAction* >(); break;
            }
            break;
        }
    } else if (_c == QMetaObject::IndexOfMethod) {
        int *result = reinterpret_cast<int *>(_a[0]);
        {
            using _t = void (MainWindow::*)();
            if (*reinterpret_cast<_t *>(_a[1]) == static_cast<_t>(&MainWindow::retranslate)) {
                *result = 0;
                return;
            }
        }
        {
            using _t = void (MainWindow::*)();
            if (*reinterpret_cast<_t *>(_a[1]) == static_cast<_t>(&MainWindow::updateMenu)) {
                *result = 1;
                return;
            }
        }
    }
}

QT_INIT_METAOBJECT const QMetaObject MainWindow::staticMetaObject = { {
    QMetaObject::SuperData::link<QMainWindow::staticMetaObject>(),
    qt_meta_stringdata_MainWindow.data,
    qt_meta_data_MainWindow,
    qt_static_metacall,
    nullptr,
    nullptr
} };


const QMetaObject *MainWindow::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *MainWindow::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_meta_stringdata_MainWindow.stringdata0))
        return static_cast<void*>(this);
    return QMainWindow::qt_metacast(_clname);
}

int MainWindow::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QMainWindow::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 61)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 61;
    } else if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 61)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 61;
    }
    return _id;
}

// SIGNAL 0
void MainWindow::retranslate()
{
    QMetaObject::activate(this, &staticMetaObject, 0, nullptr);
}

// SIGNAL 1
void MainWindow::updateMenu()
{
    QMetaObject::activate(this, &staticMetaObject, 1, nullptr);
}
struct qt_meta_stringdata_MenuAction_t {
    QByteArrayData data[4];
    char stringdata0[31];
};
#define QT_MOC_LITERAL(idx, ofs, len) \
    Q_STATIC_BYTE_ARRAY_DATA_HEADER_INITIALIZER_WITH_OFFSET(len, \
    qptrdiff(offsetof(qt_meta_stringdata_MenuAction_t, stringdata0) + ofs \
        - idx * sizeof(QByteArrayData)) \
    )
static const qt_meta_stringdata_MenuAction_t qt_meta_stringdata_MenuAction = {
    {
QT_MOC_LITERAL(0, 0, 10), // "MenuAction"
QT_MOC_LITERAL(1, 11, 11), // "retranslate"
QT_MOC_LITERAL(2, 23, 0), // ""
QT_MOC_LITERAL(3, 24, 6) // "update"

    },
    "MenuAction\0retranslate\0\0update"
};
#undef QT_MOC_LITERAL

static const uint qt_meta_data_MenuAction[] = {

 // content:
       8,       // revision
       0,       // classname
       0,    0, // classinfo
       2,   14, // methods
       0,    0, // properties
       0,    0, // enums/sets
       0,    0, // constructors
       0,       // flags
       0,       // signalCount

 // slots: name, argc, parameters, tag, flags
       1,    0,   24,    2, 0x0a /* Public */,
       3,    0,   25,    2, 0x0a /* Public */,

 // slots: parameters
    QMetaType::Void,
    QMetaType::Void,

       0        // eod
};

void MenuAction::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    if (_c == QMetaObject::InvokeMetaMethod) {
        auto *_t = static_cast<MenuAction *>(_o);
        (void)_t;
        switch (_id) {
        case 0: _t->retranslate(); break;
        case 1: _t->update(); break;
        default: ;
        }
    }
    (void)_a;
}

QT_INIT_METAOBJECT const QMetaObject MenuAction::staticMetaObject = { {
    QMetaObject::SuperData::link<QAction::staticMetaObject>(),
    qt_meta_stringdata_MenuAction.data,
    qt_meta_data_MenuAction,
    qt_static_metacall,
    nullptr,
    nullptr
} };


const QMetaObject *MenuAction::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *MenuAction::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_meta_stringdata_MenuAction.stringdata0))
        return static_cast<void*>(this);
    return QAction::qt_metacast(_clname);
}

int MenuAction::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QAction::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 2)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 2;
    } else if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 2)
            *reinterpret_cast<int*>(_a[0]) = -1;
        _id -= 2;
    }
    return _id;
}
struct qt_meta_stringdata_MenuActionGroup_t {
    QByteArrayData data[1];
    char stringdata0[16];
};
#define QT_MOC_LITERAL(idx, ofs, len) \
    Q_STATIC_BYTE_ARRAY_DATA_HEADER_INITIALIZER_WITH_OFFSET(len, \
    qptrdiff(offsetof(qt_meta_stringdata_MenuActionGroup_t, stringdata0) + ofs \
        - idx * sizeof(QByteArrayData)) \
    )
static const qt_meta_stringdata_MenuActionGroup_t qt_meta_stringdata_MenuActionGroup = {
    {
QT_MOC_LITERAL(0, 0, 15) // "MenuActionGroup"

    },
    "MenuActionGroup"
};
#undef QT_MOC_LITERAL

static const uint qt_meta_data_MenuActionGroup[] = {

 // content:
       8,       // revision
       0,       // classname
       0,    0, // classinfo
       0,    0, // methods
       0,    0, // properties
       0,    0, // enums/sets
       0,    0, // constructors
       0,       // flags
       0,       // signalCount

       0        // eod
};

void MenuActionGroup::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    (void)_o;
    (void)_id;
    (void)_c;
    (void)_a;
}

QT_INIT_METAOBJECT const QMetaObject MenuActionGroup::staticMetaObject = { {
    QMetaObject::SuperData::link<QActionGroup::staticMetaObject>(),
    qt_meta_stringdata_MenuActionGroup.data,
    qt_meta_data_MenuActionGroup,
    qt_static_metacall,
    nullptr,
    nullptr
} };


const QMetaObject *MenuActionGroup::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *MenuActionGroup::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_meta_stringdata_MenuActionGroup.stringdata0))
        return static_cast<void*>(this);
    return QActionGroup::qt_metacast(_clname);
}

int MenuActionGroup::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QActionGroup::qt_metacall(_c, _id, _a);
    return _id;
}
struct qt_meta_stringdata_MenuTree_t {
    QByteArrayData data[3];
    char stringdata0[22];
};
#define QT_MOC_LITERAL(idx, ofs, len) \
    Q_STATIC_BYTE_ARRAY_DATA_HEADER_INITIALIZER_WITH_OFFSET(len, \
    qptrdiff(offsetof(qt_meta_stringdata_MenuTree_t, stringdata0) + ofs \
        - idx * sizeof(QByteArrayData)) \
    )
static const qt_meta_stringdata_MenuTree_t qt_meta_stringdata_MenuTree = {
    {
QT_MOC_LITERAL(0, 0, 8), // "MenuTree"
QT_MOC_LITERAL(1, 9, 11), // "retranslate"
QT_MOC_LITERAL(2, 21, 0) // ""

    },
    "MenuTree\0retranslate\0"
};
#undef QT_MOC_LITERAL

static const uint qt_meta_data_MenuTree[] = {

 // content:
       8,       // revision
       0,       // classname
       0,    0, // classinfo
       1,   14, // methods
       0,    0, // properties
       0,    0, // enums/sets
       0,    0, // constructors
       0,       // flags
       0,       // signalCount

 // slots: name, argc, parameters, tag, flags
       1,    0,   19,    2, 0x0a /* Public */,

 // slots: parameters
    QMetaType::Void,

       0        // eod
};

void MenuTree::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    if (_c == QMetaObject::InvokeMetaMethod) {
        auto *_t = static_cast<MenuTree *>(_o);
        (void)_t;
        switch (_id) {
        case 0: _t->retranslate(); break;
        default: ;
        }
    }
    (void)_a;
}

QT_INIT_METAOBJECT const QMetaObject MenuTree::staticMetaObject = { {
    QMetaObject::SuperData::link<QMenu::staticMetaObject>(),
    qt_meta_stringdata_MenuTree.data,
    qt_meta_data_MenuTree,
    qt_static_metacall,
    nullptr,
    nullptr
} };


const QMetaObject *MenuTree::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *MenuTree::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_meta_stringdata_MenuTree.stringdata0))
        return static_cast<void*>(this);
    return QMenu::qt_metacast(_clname);
}

int MenuTree::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QMenu::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 1)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 1;
    } else if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 1)
            *reinterpret_cast<int*>(_a[0]) = -1;
        _id -= 1;
    }
    return _id;
}
QT_WARNING_POP
QT_END_MOC_NAMESPACE
