// Gotta define it somewhere.
#define MINIMP3_IMPLEMENTATION
#define MINIMP3_NO_STDIO

#ifdef _MSC_VER
#pragma warning(disable:4267)
#pragma warning(disable:4244)
#endif

#include "minimp3.h"
#include "minimp3_ex.h"
