#pragma once

#include "ppsspp_config.h"

#include "Common/UI/UIScreen.h"
#include "UI/BaseScreens.h"
#include "UI/TabbedDialogScreen.h"

// Per-game settings screen - enables you to configure graphic options, control options, etc
// per game.
class DriverManagerScreen : public UITabbedBaseDialogScreen {
public:
	DriverManagerScreen(const Path &gamePath);

	const char *tag() const override { return "DriverManagerScreen"; }

protected:
	void CreateTabs() override;
	bool ShowSearchControls() const override { return false; }

private:
	void OnCustomDriverInstall(UI::EventParams &e);
	void OnCustomDriverUninstall(UI::EventParams &e);
	void OnCustomDriverChange(UI::EventParams &e);

	void CreateDriverTab(UI::ViewGroup *drivers);
};
